import org.garret.jipc.*;

class QueueElement  implements java.io.Serializable {
    int    ivalue;
    String svalue;

    QueueElement(String s, int i) { 
	ivalue = i;
	svalue = s;
    }
}

/**
 * Test for queue
 * Usage:
 * 1. Start producer by "java Fifo N" command, where N is number of consumers
 * 2. Start N consumers by "java Fifo" command
 * Output: none
 */
public class Fifo extends JIPCTest { 
    static final int    nIterations = 1000;
    static final String hello = "Hello world";

    public static void main(String args[]) throws Exception { 
	JIPCSession session = createSession();
	JIPCQueue queue = session.createQueue("fifo");
	JIPCQueue result = session.createQueue("result");
	QueueElement element;

	if (args.length == 1 || (args.length == 2 && args[1].equals("0"))) { 
	    int nConsumers = Integer.parseInt(args[0], 10);
	    if (args.length == 2) { 
		nConsumers -= 1;
	    }
	    for (int i = 1; i <= nIterations; i++) { 
		queue.put(new QueueElement(hello, i));
	    }
	    // Notify consumers about termination
	    for (int i = 0; i < nConsumers; i++) { 
		queue.put(new QueueElement(hello, 0));
	    }		
	    int checksum = 0;
	    for (int i = 0; i < nConsumers; i++) { 
		element = (QueueElement)result.get();
		Assert.that(element.svalue.equals(hello));
		checksum += element.ivalue;
	    }
	    Assert.that(checksum == nIterations*(nIterations+1)/2);
	} else { 
	    int checksum = 0;
	    while (true) { 
		element = (QueueElement)queue.get();
		// System.out.println("Thread " + args[1] + " receive message " + element.ivalue);
		if (element.ivalue == 0) { 
		    break;
		}
		checksum += element.ivalue;
	    }
	    element.ivalue = checksum;
	    result.put(element);
	}
	queue.close();
	result.close();
	session.close();
    }
}
