import org.garret.jipc.*;

class SharedObject  implements java.io.Serializable {
    int    ivalue;
    String svalue;

    SharedObject(String s, int i) { 
	svalue = s;
	ivalue = i;
    }
}

/**
 * Test for mutex, barrier and shared memory
 * Usage: start N processes by "java Shmem N" command
 * Output: "Detect XXX shared object updates"
 */
public class Shmem extends JIPCTest { 
    static final int    nIterations = 1000;
    static final String hello = "Hello world";

    static class SharedMemoryMonitor extends Thread { 
	int updateCount;
	int lastValue;
	
	static final int timeout = 1000; // needed only for local sessions
	
	SharedMemoryMonitor(int lastValue) { 
	    this.lastValue = lastValue;
	}
	
        public void run() { 
	    try { 
		JIPCSession session = createSession();
		JIPCSharedMemory shmem = session.openSharedMemory("shmem");
		if (shmem != null) { 
		    while (true) {  
			if (session.isLocal()) { 
			    shmem.waitFor(timeout);	    
			} else { 
			    shmem.waitFor();	    
			}
			SharedObject so = (SharedObject)shmem.get();
			updateCount += 1;
			Assert.that(so.svalue.equals(Shmem.hello));
			if (so.ivalue == lastValue) { 
			    session.close();
			    return;
			}
		    }
		}
	    } catch(Exception x) { 
		x.printStackTrace();
	    }
	}
    }

    public static void main(String args[]) throws Exception { 
	if (args.length < 1) { 
	    System.out.println("Usage: java Shmem <number-of-processes>"); 
	    return;
	}
	SharedObject so = new SharedObject(hello, 0);

	int nProcesses = Integer.parseInt(args[0], 10);
	JIPCSession session = createSession();
	JIPCMutex mutex = session.createMutex("shmem", false);
	JIPCBarrier barrier = session.createBarrier("shmem", nProcesses);
	JIPCSharedMemory shmem = session.createSharedMemory("shmem", so);

	SharedMemoryMonitor monitor = null;
	monitor = new SharedMemoryMonitor(nProcesses*nIterations);
	monitor.start();

	barrier.waitFor();
	for (int i = 0; i < nIterations; i++) { 	    
	    mutex.lock();
	    so = (SharedObject)shmem.get();
	    so.ivalue += 1;
	    shmem.set(so);
	    mutex.unlock();
	}
	barrier.waitFor();
	so = (SharedObject)shmem.get();
	Assert.that(so.ivalue == nIterations*nProcesses);
	Assert.that(so.svalue.equals(hello));
	monitor.join();
	System.out.println("Detect " + monitor.updateCount + " shared object updates");
	session.close();
    }
}
